@@ -103,8 +103,8 @@ class Client(object):
         procid = pid_resp.json()['process_id']
         resp = self.session.post(
             self.service_url+'api/get_descriptors', 
-            data={
-                'substance_names':json.dumps(substance_names),
+            json={
+                'substance_names':substance_names,
                 'descriptor_set':descriptor_set,
                 'process_id':procid
             }
@@ -179,7 +179,7 @@ class Client(object):
         else:
             return {'success':False,'status_code':resp.status_code}
 
-    def download_descriptor_results(self,process_id,wait=False,wait_interval=1.,output_path=None):
+    def download_descriptor_results(self,process_id,wait=False,wait_interval=3.,output_path=None):
         """Get status or results of a descriptor process.
 
         This is used to download the results of a call to
@@ -228,9 +228,12 @@ class Client(object):
                     params={'process_id':process_id}
                 ).json()
         if procstat['status'] == 'FINISHED':
+            # NOTE: if it was a fractionally occupied crystal,
+            # and the job was submitted with return_files==True,
+            # this will return a zip file instead of csv!
             res = self.session.get(
                 self.service_url+'download_file',
-                data={
+                json={
                     'file_tag':'DESCRIPTORS'+str(process_id),
                     'file_ext':'.csv'
                 }
@@ -241,11 +244,17 @@ class Client(object):
             data = res._content 
             data_df = None
             try:
+                # NOTE: this should fail if the res._content is a zip file
                 data_df = pd.read_csv(BytesIO(res._content),index_col='name')
             except:
                 pass
             return {'success':True,'data':data,'df':data_df}
         else:
+            procstat['success'] = False
+            if procstat['status'] == 'ERROR':
+                procstat['error'] = 'process failed' 
+            else:
+                procstat['error'] = 'process not finished' 
             return procstat
 
 
@@ -341,7 +350,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/delete_dataset', 
-            data={'dataset_id':dataset_id}
+            json={'dataset_id':dataset_id}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -370,7 +379,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/edit_dataset', 
-            data={'dataset_id':dataset_id,'name':name,'description':description}
+            json={'dataset_id':dataset_id,'name':name,'description':description}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -396,7 +405,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/copy_dataset', 
-            data={'dataset_id':dataset_id,'name':name}
+            json={'dataset_id':dataset_id,'name':name}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -439,7 +448,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/new_model',
-            data={
+            json={
                 'dataset_id':dataset_id,
                 'name':name,
                 'description':description
@@ -469,7 +478,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/delete_model', 
-            data={'model_id':model_id}
+            json={'model_id':model_id}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -498,7 +507,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/edit_model', 
-            data={'model_id':model_id,'name':name,'description':description}
+            json={'model_id':model_id,'name':name,'description':description}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -605,7 +614,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/new_design',
-            data={
+            json={
                 'dataset_id':dataset_id,
                 'name':name,
                 'description':description
@@ -631,7 +640,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/delete_design', 
-            data={'design_id':design_id}
+            json={'design_id':design_id}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -660,7 +669,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'api/edit_design', 
-            data={'design_id':design_id,'name':name,'description':description}
+            json={'design_id':design_id,'name':name,'description':description}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -756,11 +765,11 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'configure_model', 
-            data={
+            json={
                 'model_id':model_id,
                 'output_key':output_key,
-                'scalar_columns':json.dumps(scalar_columns),
-                'substance_columns':json.dumps(substance_columns)
+                'scalar_columns':scalar_columns,
+                'substance_columns':substance_columns
             }
         )
         if resp.status_code == 200:
@@ -800,7 +809,7 @@ class Client(object):
         """
         pid_resp = self.session.get(
             self.service_url+'model_process_id', 
-            data={
+            json={
                 'model_id':model_id,
                 'process_tag':'FEATURIZATION'
             }
@@ -809,13 +818,17 @@ class Client(object):
         data = {'model_id':model_id,'process_id':procid}
         for subsk,stgs in settings.items():
             data[subsk+'_featurization'] = stgs['featurization']
-            data[subsk+'_descset'] = stgs.get('descriptor_set')
-            data[subsk+'_concwtd'] = stgs.get('concentration_weighted') or False
-            data[subsk+'_insoln'] = stgs.get('in_solution') or False
-            data[subsk+'_solnname'] = stgs.get('solution_name')
+            if 'descriptor_set' in stgs:
+                data[subsk+'_descset'] = stgs['descriptor_set']
+            if 'concentration_weighted' in stgs:
+                data[subsk+'_concwtd'] = stgs['concentration_weighted']
+            if 'in_solution' in stgs:
+                data[subsk+'_insoln'] = stgs['in_solution']
+            if 'solution_name' in stgs:
+                data[subsk+'_solnname'] = stgs['solution_name']
         resp = self.session.post(
             self.service_url+'featurize', 
-            data=data
+            json=data
         )
         if resp.status_code == 200:
             return resp.json()
@@ -847,7 +860,7 @@ class Client(object):
         """
         pid_resp = self.session.get(
             self.service_url+'model_process_id', 
-            data={
+            json={
                 'model_id':model_id,
                 'process_tag':'TRANSFORMATION'
             }
@@ -856,7 +869,7 @@ class Client(object):
         data.update(settings) 
         resp = self.session.post(
             self.service_url+'transform', 
-            data=data
+            json=data
         )
         if resp.status_code == 200:
             return resp.json()
@@ -906,18 +919,16 @@ class Client(object):
         """
         pid_resp = self.session.get(
             self.service_url+'model_process_id', 
-            data={
+            json={
                 'model_id':model_id,
                 'process_tag':'FEATURE_INSPECTION'
             }
         )
         data = {'model_id':model_id,'process_id':pid_resp.json()['process_id']}
-        if 'reserved_inputs' in settings:
-            settings['reserved_inputs'] = json.dumps(settings['reserved_inputs'])
         data.update(settings) 
         resp = self.session.post(
             self.service_url+'inspect_features', 
-            data=data
+            json=data
         )
         if resp.status_code == 200:
             return resp.json()
@@ -949,7 +960,7 @@ class Client(object):
         """
         pid_resp = self.session.get(
             self.service_url+'model_process_id', 
-            data={
+            json={
                 'model_id':model_id,
                 'process_tag':'FEATURE_SELECTION'
             }
@@ -958,7 +969,7 @@ class Client(object):
         data.update(settings) 
         resp = self.session.post(
             self.service_url+'select_features', 
-            data=data
+            json=data
         )
         if resp.status_code == 200:
             return resp.json()
@@ -1003,7 +1014,7 @@ class Client(object):
         """
         pid_resp = self.session.get(
             self.service_url+'model_process_id', 
-            data={
+            json={
                 'model_id':model_id,
                 'process_tag':'MODEL_TRAINING'
             }
@@ -1012,7 +1023,7 @@ class Client(object):
         data.update(settings) 
         resp = self.session.post(
             self.service_url+'train_model', 
-            data=data
+            json=data
         )
         if resp.status_code == 200:
             return resp.json()
@@ -1034,7 +1045,7 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'finalize_model', 
-            data={'model_id':model_id}
+            json={'model_id':model_id}
         )
         if resp.status_code == 200:
             return resp.json()
@@ -1070,7 +1081,7 @@ class Client(object):
             samples = samples.to_dict(orient='index')
         pid_resp = self.session.get(
             self.service_url+'model_process_id', 
-            data={
+            json={
                 'model_id':model_id,
                 'process_tag':'RESULTS'+uuid.uuid4().hex
             }
@@ -1078,12 +1089,12 @@ class Client(object):
         procid = pid_resp.json()['process_id']
         resp = self.session.post(
             self.service_url+'api/apply_model', 
-            data = {
+            json = {
                 'model_id':model_id,
                 'process_id':procid,
                 'run_featurization':run_featurization,
                 'run_transformation':run_transformation,
-                'samples':json.dumps(samples)
+                'samples':samples
             }
         )
         if resp.status_code == 200:
@@ -1166,11 +1177,11 @@ class Client(object):
         """
         resp = self.session.post(
             self.service_url+'configure_design', 
-            data={
+            json={
                 'design_id':design_id,
-                'scalar_columns':json.dumps(scalar_columns),
-                'substance_columns':json.dumps(substance_columns),
-                'model_output_columns':json.dumps(model_output_columns)
+                'scalar_columns':scalar_columns,
+                'substance_columns':substance_columns,
+                'model_output_columns':model_output_columns
             }
         )
         if resp.status_code == 200:
@@ -1230,7 +1241,7 @@ class Client(object):
             #for stgnm,stgval in stgs.items():
         resp = self.session.post(
             self.service_url+'generate_candidates', 
-            data=data
+            json=data
         )
         if resp.status_code == 200:
             return resp.json()
@@ -1284,7 +1295,7 @@ class Client(object):
         """
         pid_resp = self.session.get(
             self.service_url+'design_process_id', 
-            data={
+            json={
                 'design_id':design_id,
                 'process_tag':'SCREENING'
             }
@@ -1296,13 +1307,106 @@ class Client(object):
                 data[nm+'_'+stgnm] = stgval
         resp = self.session.post(
             self.service_url+'screen_candidates', 
-            data=data
+            json=data
         )
         if resp.status_code == 200:
             return resp.json()
         else:
             return {'success':False,'status_code':resp.status_code}
-            
+
+
+    # SUBSTANCES
+
+    def pull_shared_substance(self,name):
+        """Copy a substance from the shared to the private substance tables.
+
+        This allows users to pull shared substances into their private tables,
+        so that they can edit the substances without affecting the shared data,
+        and so that they can add sensitive information to their tables
+        without allowing other instances to access it.
+
+        Parameters
+        ----------
+        name : str 
+            A name that is assigned to the substance to be copied.
+
+        Returns
+        -------
+        response : dict
+            Dict containing status report and results.
+        """
+        resp = self.session.get(
+            self.service_url+'api/pull_shared_substance', 
+            json={'name':name}
+        )
+        if resp.status_code == 200:
+            respdata = resp.json()
+            return respdata
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+    def add_substance(self,name,source='',other_names=[],descriptors={},properties={},documents={}):
+        """Add a substance to the private substance tables.
+
+        See the documentation for 
+        apyonics.substances_client.SubstancesClient.add_substance().
+        """
+        resp = self.session.post(
+            self.service_url+'api/add_substance', 
+            json = {
+                'name':name,
+                'descriptors':descriptors,
+                'properties':properties,
+                'documents':documents,
+                'source':source,
+                'other_names':other_names
+            }
+        )
+        if resp.status_code == 200:
+            respdata = resp.json()
+            return respdata
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+    def update_substance(self,name,source='',other_names=[],descriptors={},properties={},documents={}):
+        """Update substance data in the private substance tables.
+
+        See the documentation for 
+        apyonics.substances_client.SubstancesClient.update_substance().
+        """
+        resp = self.session.post(
+            self.service_url+'api/update_substance', 
+            json = {
+                'name':name,
+                'source':source,
+                'other_names':other_names,
+                'descriptors':descriptors,
+                'properties':properties,
+                'documents':documents
+            }
+        )
+        if resp.status_code == 200:
+            respdata = resp.json()
+            return respdata
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
+    def get_substance_data(self,name):
+        """Get all data for a substance by providing a substance name.
+
+        See the documentation for
+        apyonics.substances_client.SubstancesClient.get_substance_data().
+        """
+        resp = self.session.get(
+            self.service_url+'api/download_substance', 
+            json={'name':name}
+        )
+        if resp.status_code == 200:
+            respdata = {'success':True,'data':resp.json()['data']}
+            return resp.json()
+        else:
+            return {'success':False,'status_code':resp.status_code}
+
 
     # UTILITIES 
 
